#include <stdio.h>
#include <stdlib.h>
#include <limits.h>

typedef enum {
	false,
	true
} bool;

static void fatal(const char *message) {
	fprintf(stderr, "error: %s\n", message);
	exit(EXIT_FAILURE);
}


//////////////////////////////////////////////////////////////////////////////
// handle reading and writing of 24-bit BMP files
//////////////////////////////////////////////////////////////////////////////
#define WIN32_LEAN_AND_MEAN
#define WIN32_EXTRA_LEAN
#include <windows.h>

typedef struct {
	unsigned char b, g, r;
} rgb_t;

typedef struct {
	unsigned xres, yres;
	rgb_t *data;
} bitmap_t;

#pragma pack(push, 1)
typedef struct {
	BITMAPFILEHEADER file;
	BITMAPINFOHEADER info;
} bmp_header;
#pragma pack(pop)

/*
static void reverse_bmp(bitmap_t bmp) {
	unsigned i;

	rgb_t
		*low = bmp.data,
		*high = &bmp.data[bmp.xres * bmp.yres];

	while(low < high) {
		high -= bmp.xres;

		for(i = 0; i < bmp.xres; ++i) {
			rgb_t swap = low[i];
			low[i] = high[i];
			high[i] = swap;
		}

		low += bmp.xres;
	}
}
*/

static bitmap_t load_bmp(const char *name) {
	FILE *file;
	bmp_header hdr;
	unsigned size;
	bitmap_t bmp;

	if(file = fopen(name, "rb"), !file) {
		fatal("unable to open bmp file");
	}

	if (
		fread(&hdr, sizeof hdr, 1, file) != 1 ||
		hdr.file.bfType != 'MB'||
		hdr.info.biPlanes != 1 ||
		hdr.info.biBitCount != 24 ||
		hdr.info.biCompression != BI_RGB
	) {
		fatal("invalid bitmap header");
	}

	size = hdr.info.biWidth * hdr.info.biHeight;;

	bmp.xres = hdr.info.biWidth;
	bmp.yres = hdr.info.biHeight;
	bmp.data = malloc(size * sizeof(rgb_t));

	fseek(file, hdr.file.bfOffBits, SEEK_SET);

	{
		unsigned y;

		rgb_t *p = &bmp.data[size];

		for(y = 0; y < bmp.yres; ++y) {
			p -= bmp.xres;

			if(fread(p, bmp.xres * sizeof(rgb_t), 1, file) != 1) {
				fatal("unable to read bitmap data");
			}
		}
	}

	fclose(file);

	return bmp;
}

static void save_bmp(bitmap_t bmp, const char *name) {
	FILE *file;

	unsigned size = bmp.xres * bmp.yres;

	bmp_header hdr = {
		{
			'MB',
			sizeof(bmp_header) + size,
			0,
			0,
			sizeof(bmp_header),
		},

		{
			sizeof(BITMAPINFOHEADER),
			bmp.xres,
			bmp.yres,
			1,
			24,
			BI_RGB,
			size * sizeof(rgb_t),
			0,
			0,
			0,
			0
		}
	};

	if(file = fopen(name, "wb"), !file) {
		fatal("unable to create bmp file");
	}

	fwrite(&hdr, sizeof hdr, 1, file);

	{
		unsigned y;

		const rgb_t *p = &bmp.data[size];

		for(y = 0; y < bmp.yres; ++y) {
			p -= bmp.xres;
			fwrite(p, bmp.xres * sizeof(rgb_t), 1, file);
		}
	}

	fclose(file);
}


//////////////////////////////////////////////////////////////////////////////
// map rgb colors to/from the vic palette
//////////////////////////////////////////////////////////////////////////////
#define BGR(r, g, b) { b, g, r }

static const rgb_t vic_palette[] = {
#	if 1
	// GangEd
	BGR(0x00, 0x00, 0x00),
	BGR(0xff, 0xff, 0xff),
	BGR(0x89, 0x40, 0x36),
	BGR(0x7a, 0xbf, 0xc7),
	BGR(0x8a, 0x46, 0xae),
	BGR(0x68, 0xa9, 0x41),
	BGR(0x3e, 0x31, 0xa2),
	BGR(0xd0, 0xdc, 0x71),
	BGR(0x90, 0x5f, 0x25),
	BGR(0x5c, 0x47, 0x00),
	BGR(0xbb, 0x77, 0x6d),
	BGR(0x55, 0x55, 0x55),
	BGR(0x80, 0x80, 0x80),
	BGR(0xac, 0xea, 0x88),
	BGR(0x7c, 0x70, 0xda),
	BGR(0xab, 0xab, 0xab),
#	elif 0
	// Deekay
	BGR(0x00, 0x00, 0x00),
	BGR(0xff, 0xff, 0xff),
	BGR(0x68, 0x37, 0x2b),
	BGR(0x70, 0xa4, 0xb2),
	BGR(0x6f, 0x3d, 0x86),
	BGR(0x58, 0x8d, 0x43),
	BGR(0x35, 0x28, 0x79),
	BGR(0xb8, 0xc7, 0x6f),
	BGR(0x6f, 0x4f, 0x25),
	BGR(0x43, 0x39, 0x00),
	BGR(0x9a, 0x67, 0x59),
	BGR(0x44, 0x44, 0x44),
	BGR(0x6c, 0x6c, 0x6c),
	BGR(0x9a, 0xd2, 0x84),
	BGR(0x6c, 0x5e, 0xb5),
	BGR(0x95, 0x95, 0x95)
#	else
	// VICE
	BGR(0x00, 0x00, 0x00),
	BGR(0xfd, 0xfe, 0xfc),
	BGR(0xbe, 0x1a, 0x24),
	BGR(0x30, 0xe6, 0xc6),
	BGR(0xb4, 0x1a, 0xe2),
	BGR(0x1f, 0xd2, 0x1e),
	BGR(0x21, 0x1b, 0xae),
	BGR(0xdf, 0xf6, 0x0a),
	BGR(0xb8, 0x41, 0x04),
	BGR(0x6a, 0x33, 0x04),
	BGR(0xfe, 0x4a, 0x57),
	BGR(0x42, 0x45, 0x40),
	BGR(0x70, 0x74, 0x6f),
	BGR(0x59, 0xfe, 0x59),
	BGR(0x5f, 0x53, 0xfe),
	BGR(0xa4, 0xa7, 0xa2)
#	endif
};

static char vic_nearest(rgb_t color) {
	char best_index;
	unsigned best_match = UINT_MAX;

	static struct {
		rgb_t color;
		char index;
	} cache[4];

	char i;

	for(i = 0; i < 4; ++i) {
		if (
			color.r == cache[i].color.r &&
			color.g == cache[i].color.g &&
			color.b == cache[i].color.b
		) {
			return cache[i].index;
		}
	}

	for(i = 0; i < 16; ++i) {
		unsigned distance;

		signed
			r = color.r - vic_palette[i].r,
			g = color.g - vic_palette[i].g,
			b = color.b - vic_palette[i].b;

		r *= r;
		g *= g;
		b *= b;

		r *= 299;
		g *= 587;
		b *= 114;

		distance = r + g + b;

		if(distance < best_match) {
			best_index = i;
			best_match = distance;
		}
	}

	cache[3] = cache[2];
	cache[2] = cache[1];
	cache[1] = cache[0];
	cache[0].color = color;
	cache[0].index = best_index;

	return best_index;
}


//////////////////////////////////////////////////////////////////////////////
// try to fit the colors in an 8x16 bitmap chunk to their final C64
// bit-patterns, this involves selecting a per-char color and choosing between
// hires and multicolor
//////////////////////////////////////////////////////////////////////////////
enum {
	TILE_XRES = 8,
	TILE_YRES = 16,
	TILE_UNASSIGNED_COLOR = 16,
	TILE_IMPOSSIBLE_COLOR = 17
};

static char try_hires(const rgb_t *src, unsigned char *dst, unsigned pitch,
	char background) {
	unsigned x, y;

	char colorram = TILE_UNASSIGNED_COLOR;

	for(y = 0; y < TILE_YRES; ++y) {
		unsigned char pattern = 0;

		for(x = 0; x < TILE_XRES; ++x) {
			char color = vic_nearest(src[x]);

			pattern <<= 1;

			if(color != background) {
				pattern |= 1;

				if(color != colorram) {
					if(colorram != TILE_UNASSIGNED_COLOR || color >= 8) {
						return TILE_IMPOSSIBLE_COLOR;
					}

					colorram = color;
				}
			}
		}

		src += pitch;
		*dst++ = pattern;
	}

	return colorram;
}

static char try_multicolor(const rgb_t *src, unsigned char *dst, unsigned pitch,
	char background, char multicolor1, char multicolor2) {
	unsigned x, y;

	char colorram = TILE_UNASSIGNED_COLOR;

	for(y = 0; y < TILE_YRES; ++y) {
		unsigned char pattern = 0;

		for(x = 0; x < TILE_XRES; x += 2) {
			char color = vic_nearest(src[x]);

			if(vic_nearest(src[x + 1]) != color) {
				return TILE_IMPOSSIBLE_COLOR;
			}

			pattern <<= 2;

			if(color == multicolor1) {
				pattern |= 1;
			} else if(color == multicolor2) {
				pattern |= 2;
			} else if(color != background) {
				pattern |= 3;

				if(color != colorram) {
					if(colorram != TILE_UNASSIGNED_COLOR || color >= 8) {
						return TILE_IMPOSSIBLE_COLOR;
					}

					colorram = color;
				}
			}
		}

		src += pitch;
		*dst++ = pattern;
	}

	return colorram | 8;
}

static char grab_tile(const rgb_t *src, unsigned char *dst, unsigned pitch,
	char background, char multicolor1, char multicolor2) {
	char color = try_hires(src, dst, pitch, background);

	if(color == TILE_IMPOSSIBLE_COLOR) {
		color = try_multicolor(src, dst, pitch, background, multicolor1,
			multicolor2);

		if(color == TILE_IMPOSSIBLE_COLOR) {
			memset(dst, 0, TILE_YRES);
		}
	}

	return color;
}

static void render_tile(const unsigned char *src, rgb_t *dst, unsigned pitch,
	char background, char multicolor1, char multicolor2, char colorram) {
	unsigned x, y;

	if(colorram < 8) {
		// hires
		for(y = 0; y < TILE_YRES; ++y) {
			unsigned char pattern = *src++;

			for(x = 0; x < TILE_XRES; ++x) {
				char color = (pattern >> 7) ? colorram : background;
				dst[x] = vic_palette[color];
				pattern <<= 1;
			}

			dst += pitch;
		}
	} else {
		// multicolor
		colorram &= 7;

		for(y = 0; y < TILE_YRES; ++y) {
			unsigned char pattern = *src++;

			for(x = 0; x < TILE_XRES; x += 2) {
				char color;

				switch(pattern >> 6) {
				case 0: color = background; break;
				case 1: color = multicolor1; break;
				case 2: color = multicolor2; break;
				case 3: color = colorram; break;
				}

				dst[x + 0] = dst[x + 1] = vic_palette[color];
				pattern <<= 2;
			}

			dst += pitch;
		}
	}
}


//////////////////////////////////////////////////////////////////////////////
// keep track of the tileset and detect any duplicates
//////////////////////////////////////////////////////////////////////////////
enum { NULL_TILE = -1 };

static struct {
	unsigned used;
	signed bucket[256];
	signed link[256];
	unsigned char data[256][TILE_YRES];
} tileset;

static void tileset_reset(void) {
	unsigned i;

	for(i = 0; i < 256; ++i) {
		tileset.bucket[i] = NULL_TILE;
	}

	tileset.used = 0;
}

static signed tileset_add(unsigned char *tile, bool *reused) {
	unsigned y;
	signed ptr;

	unsigned char key = 0;

	for(y = 0; y < TILE_YRES; ++y) {
		key ^= tile[y];
	}

	for(ptr = tileset.bucket[key]; ptr != NULL_TILE; ptr = tileset.link[ptr]) {
		if(!memcmp(tileset.data[ptr], tile, TILE_YRES)) {
			*reused = true;
			return ptr;
		}
	}

	if(tileset.used >= 256) {
		return NULL_TILE;
	}

	memcpy(tileset.data[tileset.used], tile, TILE_YRES);

	tileset.link[tileset.used] = tileset.bucket[key];
	tileset.bucket[key] = tileset.used;

	*reused = false;
	return tileset.used++;
}


//////////////////////////////////////////////////////////////////////////////
// manage tilesheets
//////////////////////////////////////////////////////////////////////////////
typedef struct {
	signed index;
	char color;
} tileinfo_t ;

static struct {
	tileinfo_t *tiles;

	unsigned xcount;
	unsigned ycount;
	unsigned count;

	char background;
	char multicolor1;
	char multicolor2;
} tilesheet;

static char get_color_cell(rgb_t *bitmap, unsigned pitch) {
	char color = vic_nearest(*bitmap);
	rgb_t rgb = vic_palette[color];

	unsigned x, y;

	for(y = 0; y < TILE_YRES; ++y) {
		for(x = 0; x < TILE_XRES; ++x) {
			bitmap[x] = rgb;
		}

		bitmap += pitch;
	}

	return color;
}

static void shade_cell(rgb_t *bitmap, unsigned pitch, rgb_t color) {
	unsigned x, y;

	color.r >>= 1;
	color.g >>= 1;
	color.b >>= 1;

	for(y = 0; y < TILE_YRES; ++y) {
		for(x = 0; x < TILE_XRES; ++x) {
			bitmap[x].r = (bitmap[x].r >> 1) + color.r;
			bitmap[x].g = (bitmap[x].g >> 1) + color.g;
			bitmap[x].b = (bitmap[x].b >> 1) + color.b;
		}

		bitmap += pitch;
	}
}

static tileinfo_t add_to_tilesheet(rgb_t *bitmap, unsigned pitch, char background,
	char multicolor1, char multicolor2) {
	static const rgb_t
		red_shade = BGR(0xff, 0x00, 0x00),
		black_shade = BGR(0x00, 0x00, 0x00),
		blue_shade = BGR(0x00, 0x00, 0xff);

	tileinfo_t info;
	unsigned char tile[TILE_YRES];
	bool reused;

	info.color = grab_tile(bitmap, tile, pitch, background, multicolor1, multicolor2);
	info.index = tileset_add(tile, &reused);

	/*
	{
		unsigned x, y;

		rgb_t *p = bitmap;

		for(y = 0; y < TILE_YRES; ++y) {
			for(x = 0; x < TILE_XRES; ++x) {
				p[x] = vic_palette[vic_nearest(p[x])];
			}

			p += pitch;
		}
	}
	*/

	if(info.color == TILE_IMPOSSIBLE_COLOR) {
		shade_cell(bitmap, pitch, red_shade);
	} else if(info.index == NULL_TILE) {
		shade_cell(bitmap, pitch, black_shade); 
	} else {
		render_tile(tileset.data[info.index], bitmap, pitch, background, multicolor1,
			multicolor2, info.color);
	
		if(reused) {
			shade_cell(bitmap, pitch, blue_shade);
		}
	}

	return info;
}

static void grab_tilesheet(bitmap_t bmp) {
	rgb_t *bitmap;
	unsigned x, y;
	tileinfo_t *tile;

	// verify the bitmap's dimensions
	if(bmp.xres % TILE_XRES || bmp.yres % TILE_YRES) {
		fatal("the bitmap's size must be an even multiple of 8x16");
	}

	tilesheet.xcount = bmp.xres / TILE_XRES;
	tilesheet.ycount = bmp.yres / TILE_YRES;

	if(tilesheet.xcount < 3 || tilesheet.ycount-- < 1) {
		fatal("the bitmap is too small");
	}

	tilesheet.count = tilesheet.xcount * tilesheet.ycount;
	tilesheet.tiles = malloc(tilesheet.count * sizeof(tileinfo_t));

	// get the fixed char colors
	bitmap = bmp.data;

	{
		rgb_t *bottom = 
			&bitmap[tilesheet.xcount * TILE_XRES * tilesheet.ycount * TILE_YRES];

		tilesheet.background = get_color_cell(bottom, bmp.xres);
		bottom += TILE_XRES;
		tilesheet.multicolor1 = get_color_cell(bottom, bmp.xres);
		bottom += TILE_XRES;
		tilesheet.multicolor2 = get_color_cell(bottom, bmp.xres);
	}

	// and convert each tile on the sheet
	tileset_reset();

	tile = tilesheet.tiles;

	for(y = 0; y < tilesheet.ycount; ++y) {
		for(x = 0; x < tilesheet.xcount; ++x) {
			*tile++ = add_to_tilesheet(&bitmap[x * TILE_XRES], bmp.xres, tilesheet.background,
				tilesheet.multicolor1, tilesheet.multicolor2);
		}

		bitmap += bmp.xres * TILE_YRES;
	}
}

static void render_tilesheet(bitmap_t bmp) {
	unsigned x, y;

	const tileinfo_t *src = tilesheet.tiles;
	rgb_t *dst = bmp.data;

	for(y = 0; y < tilesheet.ycount; ++y) {
		for(x = 0; x < tilesheet.xcount; ++x) {
			render_tile(tileset.data[src->index], &dst[x * TILE_XRES], bmp.xres,
				tilesheet.background, tilesheet.multicolor1, tilesheet.multicolor2, src->color);

			++src;
		}

		dst += bmp.xres * TILE_YRES;
	}
}

static void save_tilesheet(const char *name) {
	FILE *file;

	if(file = fopen(name, "wb"), !file) {
		fatal("unable to create tilesheet file");
	}

	// tiles
	if(TILE_YRES == 16) {
		unsigned i, j;
	
		unsigned padding = (256 - tileset.used) << 3;

		for(i = 0; i < tileset.used; ++i) {
			for(j = 0; j < 4; ++j) {
				fputc(tileset.data[i][j], file);
			}

			for(j = 0; j < 4; ++j) {
				fputc(tileset.data[i][12 + j], file);
			}
		}

		fseek(file, padding, SEEK_CUR);

		for(i = 0; i < tileset.used; ++i) {
			for(j = 0; j < 4; ++j) {
				fputc(tileset.data[i][8 + j], file);
			}

			for(j = 0; j < 4; ++j) {
				fputc(tileset.data[i][4 + j], file);
			}
		}

		fseek(file, padding, SEEK_CUR);
	} else {
		fwrite(tileset.data, tileset.used, TILE_YRES, file);
	}

	// static palette
	fputc(tilesheet.background, file);
	fputc(tilesheet.multicolor1, file);
	fputc(tilesheet.multicolor2, file);

	// tile indices and colors
	{
		unsigned x, y;

		// write in reverse order, to accommodate upwards scrolling
		const tileinfo_t *p = &tilesheet.tiles[tilesheet.count];

		for(y = 0; y < tilesheet.ycount; ++y) {
			p -= tilesheet.xcount;

			for(x = 0; x < tilesheet.xcount; ++x) {
				fputc(p[x].index, file);
				fputc(p[x].color, file);
			}
		}
	}

	fclose(file);
}


//////////////////////////////////////////////////////////////////////////////
// the main thing of the main thing is to do the main thing
//////////////////////////////////////////////////////////////////////////////
int main(int argc, char *argv[]) {
	char *f_input;
	char *f_extension;
	char f_output[256];
	
	bitmap_t bmp;

	if(argc != 2) {
		fatal("syntax: tilesheet.exe {input.bmp}");
	}

	f_input = argv[1];
	bmp = load_bmp(f_input);
	
	if(f_extension = strrchr(f_input, '.'), f_extension) {
		*f_extension = '\0';
	}

	grab_tilesheet(bmp);
	sprintf(f_output, "%s_diag.bmp", f_input);

	save_bmp(bmp, f_output);

	render_tilesheet(bmp);
	sprintf(f_output, "%s_sheet.bmp", f_input);

	save_bmp(bmp, f_output);

	sprintf(f_output, "%s.bin", f_input);
	save_tilesheet(f_output);

	printf("%d unique tiles (%g%%)\n", tileset.used, tileset.used / 2.56);
	return EXIT_SUCCESS;
}